@@ -1,7 +1,7 @@
 // args.rs
 
 // *************************************************************************
-// * Copyright (C) 2018 Daniel Mueller (deso@posteo.net)                   *
+// * Copyright (C) 2018-2019 Daniel Mueller (deso@posteo.net)              *
 // *                                                                       *
 // * This program is free software: you can redistribute it and/or modify  *
 // * it under the terms of the GNU General Public License as published by  *
@@ -17,7 +17,7 @@
 // * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 // *************************************************************************
 
-use std::fmt;
+use std::ffi;
 use std::io;
 use std::result;
 use std::str;
@@ -25,107 +25,81 @@ use std::str;
 use crate::commands;
 use crate::error::Error;
 use crate::pinentry;
+use crate::RunCtx;
 
 type Result<T> = result::Result<T, Error>;
 
-/// A top-level command for nitrocli.
-#[derive(Debug)]
-pub enum Command {
-  Config,
-  Lock,
-  Otp,
-  Pin,
-  Pws,
-  Status,
-  Storage,
+trait Stdio {
+  fn stdio(&mut self) -> (&mut dyn io::Write, &mut dyn io::Write);
 }
 
+impl<'io> Stdio for RunCtx<'io> {
+  fn stdio(&mut self) -> (&mut dyn io::Write, &mut dyn io::Write) {
+    (self.stdout, self.stderr)
+  }
+}
+
+/// A command execution context that captures additional data pertaining
+/// the command execution.
+pub struct ExecCtx<'io> {
+  pub model: Option<DeviceModel>,
+  pub stdout: &'io mut dyn io::Write,
+  pub stderr: &'io mut dyn io::Write,
+  pub admin_pin: Option<ffi::OsString>,
+  pub user_pin: Option<ffi::OsString>,
+  pub new_admin_pin: Option<ffi::OsString>,
+  pub new_user_pin: Option<ffi::OsString>,
+  pub verbosity: u64,
+}
+
+impl<'io> Stdio for ExecCtx<'io> {
+  fn stdio(&mut self) -> (&mut dyn io::Write, &mut dyn io::Write) {
+    (self.stdout, self.stderr)
+  }
+}
+
+/// The available Nitrokey models.
+Enum! {DeviceModel, [
+  Pro => "pro",
+  Storage => "storage"
+]}
+
+/// A top-level command for nitrocli.
+Enum! {Command, [
+  Config => "config",
+  Lock => "lock",
+  Otp => "otp",
+  Pin => "pin",
+  Pws => "pws",
+  Status => "status",
+  Storage => "storage"
+]}
+
 impl Command {
   /// Execute this command with the given arguments.
-  pub fn execute(&self, args: Vec<String>) -> Result<()> {
-    match *self {
-      Command::Config => config(args),
-      Command::Lock => lock(args),
-      Command::Otp => otp(args),
-      Command::Pin => pin(args),
-      Command::Pws => pws(args),
-      Command::Status => status(args),
-      Command::Storage => storage(args),
+  pub fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+    match self {
+      Command::Config => config(ctx, args),
+      Command::Lock => lock(ctx, args),
+      Command::Otp => otp(ctx, args),
+      Command::Pin => pin(ctx, args),
+      Command::Pws => pws(ctx, args),
+      Command::Status => status(ctx, args),
+      Command::Storage => storage(ctx, args),
     }
   }
 }
 
-impl fmt::Display for Command {
-  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-    write!(
-      f,
-      "{}",
-      match *self {
-        Command::Config => "config",
-        Command::Lock => "lock",
-        Command::Otp => "otp",
-        Command::Pin => "pin",
-        Command::Pws => "pws",
-        Command::Status => "status",
-        Command::Storage => "storage",
-      }
-    )
-  }
-}
-
-impl str::FromStr for Command {
-  type Err = ();
-
-  fn from_str(s: &str) -> result::Result<Self, Self::Err> {
-    match s {
-      "config" => Ok(Command::Config),
-      "lock" => Ok(Command::Lock),
-      "otp" => Ok(Command::Otp),
-      "pin" => Ok(Command::Pin),
-      "pws" => Ok(Command::Pws),
-      "status" => Ok(Command::Status),
-      "storage" => Ok(Command::Storage),
-      _ => Err(()),
-    }
-  }
-}
-
-#[derive(Debug)]
-enum ConfigCommand {
-  Get,
-  Set,
-}
+Enum! {ConfigCommand, [
+  Get => "get",
+  Set => "set"
+]}
 
 impl ConfigCommand {
-  fn execute(&self, args: Vec<String>) -> Result<()> {
-    match *self {
-      ConfigCommand::Get => config_get(args),
-      ConfigCommand::Set => config_set(args),
-    }
-  }
-}
-
-impl fmt::Display for ConfigCommand {
-  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-    write!(
-      f,
-      "{}",
-      match *self {
-        ConfigCommand::Get => "get",
-        ConfigCommand::Set => "set",
-      }
-    )
-  }
-}
-
-impl str::FromStr for ConfigCommand {
-  type Err = ();
-
-  fn from_str(s: &str) -> result::Result<Self, Self::Err> {
-    match s {
-      "get" => Ok(ConfigCommand::Get),
-      "set" => Ok(ConfigCommand::Set),
-      _ => Err(()),
+  fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+    match self {
+      ConfigCommand::Get => config_get(ctx, args),
+      ConfigCommand::Set => config_set(ctx, args),
     }
   }
 }
@@ -165,115 +139,33 @@ impl<T> ConfigOption<T> {
   }
 }
 
-#[derive(Debug)]
-enum OtpCommand {
-  Clear,
-  Get,
-  Set,
-  Status,
-}
+Enum! {OtpCommand, [
+  Clear => "clear",
+  Get => "get",
+  Set => "set",
+  Status => "status"
+]}
 
 impl OtpCommand {
-  fn execute(&self, args: Vec<String>) -> Result<()> {
-    match *self {
-      OtpCommand::Clear => otp_clear(args),
-      OtpCommand::Get => otp_get(args),
-      OtpCommand::Set => otp_set(args),
-      OtpCommand::Status => otp_status(args),
+  fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+    match self {
+      OtpCommand::Clear => otp_clear(ctx, args),
+      OtpCommand::Get => otp_get(ctx, args),
+      OtpCommand::Set => otp_set(ctx, args),
+      OtpCommand::Status => otp_status(ctx, args),
     }
   }
 }
 
-impl fmt::Display for OtpCommand {
-  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-    write!(
-      f,
-      "{}",
-      match *self {
-        OtpCommand::Clear => "clear",
-        OtpCommand::Get => "get",
-        OtpCommand::Set => "set",
-        OtpCommand::Status => "status",
-      }
-    )
-  }
-}
+Enum! {OtpAlgorithm, [
+  Hotp => "hotp",
+  Totp => "totp"
+]}
 
-impl str::FromStr for OtpCommand {
-  type Err = ();
-
-  fn from_str(s: &str) -> result::Result<Self, Self::Err> {
-    match s {
-      "clear" => Ok(OtpCommand::Clear),
-      "get" => Ok(OtpCommand::Get),
-      "set" => Ok(OtpCommand::Set),
-      "status" => Ok(OtpCommand::Status),
-      _ => Err(()),
-    }
-  }
-}
-
-#[derive(Clone, Copy, Debug, PartialEq)]
-pub enum OtpAlgorithm {
-  Hotp,
-  Totp,
-}
-
-impl fmt::Display for OtpAlgorithm {
-  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-    write!(
-      f,
-      "{}",
-      match *self {
-        OtpAlgorithm::Hotp => "hotp",
-        OtpAlgorithm::Totp => "totp",
-      }
-    )
-  }
-}
-
-impl str::FromStr for OtpAlgorithm {
-  type Err = ();
-
-  fn from_str(s: &str) -> result::Result<Self, Self::Err> {
-    match s {
-      "hotp" => Ok(OtpAlgorithm::Hotp),
-      "totp" => Ok(OtpAlgorithm::Totp),
-      _ => Err(()),
-    }
-  }
-}
-
-#[derive(Clone, Copy, Debug)]
-enum OtpMode {
-  SixDigits,
-  EightDigits,
-}
-
-impl fmt::Display for OtpMode {
-  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-    write!(
-      f,
-      "{}",
-      match *self {
-        OtpMode::SixDigits => "6",
-        OtpMode::EightDigits => "8",
-      }
-    )
-  }
-}
-
-impl str::FromStr for OtpMode {
-  type Err = ();
-
-  fn from_str(s: &str) -> result::Result<Self, Self::Err> {
-    match s {
-      "6" => Ok(OtpMode::SixDigits),
-      "8" => Ok(OtpMode::EightDigits),
-      _ => Err(()),
-    }
-  }
-}
+Enum! {OtpMode, [
+  SixDigits => "6",
+  EightDigits => "8"
+]}
 
 impl From<OtpMode> for nitrokey::OtpMode {
   fn from(mode: OtpMode) -> Self {
@@ -284,100 +176,53 @@ impl From<OtpMode> for nitrokey::OtpMode {
   }
 }
 
-#[derive(Debug)]
-enum PinCommand {
-  Clear,
-  Set,
-  Unblock,
-}
+Enum! {OtpSecretFormat, [
+  Ascii => "ascii",
+  Base32 => "base32",
+  Hex => "hex"
+]}
+
+Enum! {PinCommand, [
+  Clear => "clear",
+  Set => "set",
+  Unblock => "unblock"
+]}
 
 impl PinCommand {
-  fn execute(&self, args: Vec<String>) -> Result<()> {
-    match *self {
-      PinCommand::Clear => pin_clear(args),
-      PinCommand::Set => pin_set(args),
-      PinCommand::Unblock => pin_unblock(args),
+  fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+    match self {
+      PinCommand::Clear => pin_clear(ctx, args),
+      PinCommand::Set => pin_set(ctx, args),
+      PinCommand::Unblock => pin_unblock(ctx, args),
     }
   }
 }
 
-impl fmt::Display for PinCommand {
-  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-    write!(
-      f,
-      "{}",
-      match *self {
-        PinCommand::Clear => "clear",
-        PinCommand::Set => "set",
-        PinCommand::Unblock => "unblock",
-      }
-    )
-  }
-}
-
-impl str::FromStr for PinCommand {
-  type Err = ();
-
-  fn from_str(s: &str) -> result::Result<Self, Self::Err> {
-    match s {
-      "clear" => Ok(PinCommand::Clear),
-      "set" => Ok(PinCommand::Set),
-      "unblock" => Ok(PinCommand::Unblock),
-      _ => Err(()),
-    }
-  }
-}
-
-#[derive(Debug)]
-enum PwsCommand {
-  Clear,
-  Get,
-  Set,
-  Status,
-}
+Enum! {PwsCommand, [
+  Clear => "clear",
+  Get => "get",
+  Set => "set",
+  Status => "status"
+]}
 
 impl PwsCommand {
-  fn execute(&self, args: Vec<String>) -> Result<()> {
-    match *self {
-      PwsCommand::Clear => pws_clear(args),
-      PwsCommand::Get => pws_get(args),
-      PwsCommand::Set => pws_set(args),
-      PwsCommand::Status => pws_status(args),
+  fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+    match self {
+      PwsCommand::Clear => pws_clear(ctx, args),
+      PwsCommand::Get => pws_get(ctx, args),
+      PwsCommand::Set => pws_set(ctx, args),
+      PwsCommand::Status => pws_status(ctx, args),
     }
   }
 }
 
-impl fmt::Display for PwsCommand {
-  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-    write!(
-      f,
-      "{}",
-      match *self {
-        PwsCommand::Clear => "clear",
-        PwsCommand::Get => "get",
-        PwsCommand::Set => "set",
-        PwsCommand::Status => "status",
-      }
-    )
-  }
-}
-
-impl str::FromStr for PwsCommand {
-  type Err = ();
-
-  fn from_str(s: &str) -> result::Result<Self, Self::Err> {
-    match s {
-      "clear" => Ok(PwsCommand::Clear),
-      "get" => Ok(PwsCommand::Get),
-      "set" => Ok(PwsCommand::Set),
-      "status" => Ok(PwsCommand::Status),
-      _ => Err(()),
-    }
-  }
-}
-
-fn parse(parser: &argparse::ArgumentParser<'_>, args: Vec<String>) -> Result<()> {
-  if let Err(err) = parser.parse(args, &mut io::stdout(), &mut io::stderr()) {
+fn parse(
+  ctx: &mut impl Stdio,
+  parser: &argparse::ArgumentParser<'_>,
+  args: Vec<String>,
+) -> Result<()> {
+  let (stdout, stderr) = ctx.stdio();
+  if let Err(err) = parser.parse(args, stdout, stderr) {
     Err(Error::ArgparseError(err))
   } else {
     Ok(())
@@ -385,60 +230,32 @@ fn parse(parser: &argparse::ArgumentParser<'_>, args: Vec<String>) -> Result<()>
 }
 
 /// Inquire the status of the nitrokey.
-fn status(args: Vec<String>) -> Result<()> {
+fn status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Prints the status of the connected Nitrokey device");
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
 
-  commands::status()
+  commands::status(ctx)
 }
 
-#[derive(Debug)]
-enum StorageCommand {
-  Close,
-  Open,
-  Status,
-}
+Enum! {StorageCommand, [
+  Close => "close",
+  Open => "open",
+  Status => "status"
+]}
 
 impl StorageCommand {
-  fn execute(&self, args: Vec<String>) -> Result<()> {
-    match *self {
-      StorageCommand::Close => storage_close(args),
-      StorageCommand::Open => storage_open(args),
-      StorageCommand::Status => storage_status(args),
-    }
-  }
-}
-
-impl fmt::Display for StorageCommand {
-  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-    write!(
-      f,
-      "{}",
-      match *self {
-        StorageCommand::Close => "close",
-        StorageCommand::Open => "open",
-        StorageCommand::Status => "status",
-      }
-    )
-  }
-}
-
-impl str::FromStr for StorageCommand {
-  type Err = ();
-
-  fn from_str(s: &str) -> result::Result<Self, Self::Err> {
-    match s {
-      "close" => Ok(StorageCommand::Close),
-      "open" => Ok(StorageCommand::Open),
-      "status" => Ok(StorageCommand::Status),
-      _ => Err(()),
+  fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+    match self {
+      StorageCommand::Close => storage_close(ctx, args),
+      StorageCommand::Open => storage_open(ctx, args),
+      StorageCommand::Status => storage_status(ctx, args),
     }
   }
 }
 
 /// Execute a storage subcommand.
-fn storage(args: Vec<String>) -> Result<()> {
+fn storage(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut subcommand = StorageCommand::Open;
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
@@ -454,42 +271,42 @@ fn storage(args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  subargs.insert(0, format!("nitrocli storage {}", subcommand));
-  subcommand.execute(subargs)
+  subargs.insert(0, format!("nitrocli {} {}", Command::Storage, subcommand));
+  subcommand.execute(ctx, subargs)
 }
 
 /// Open the encrypted volume on the nitrokey.
-fn storage_open(args: Vec<String>) -> Result<()> {
+fn storage_open(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Opens the encrypted volume on a Nitrokey Storage");
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
 
-  commands::storage_open()
+  commands::storage_open(ctx)
 }
 
 /// Close the previously opened encrypted volume.
-fn storage_close(args: Vec<String>) -> Result<()> {
+fn storage_close(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Closes the encrypted volume on a Nitrokey Storage");
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
 
-  commands::storage_close()
+  commands::storage_close(ctx)
 }
 
 /// Print the status of the nitrokey's storage.
-fn storage_status(args: Vec<String>) -> Result<()> {
+fn storage_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Prints the status of the Nitrokey's storage");
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
 
-  commands::storage_status()
+  commands::storage_status(ctx)
 }
 
 /// Execute a config subcommand.
-fn config(args: Vec<String>) -> Result<()> {
+fn config(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut subcommand = ConfigCommand::Get;
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
@@ -505,24 +322,24 @@ fn config(args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  subargs.insert(0, format!("nitrocli config {}", subcommand));
-  subcommand.execute(subargs)
+  subargs.insert(0, format!("nitrocli {} {}", Command::Config, subcommand));
+  subcommand.execute(ctx, subargs)
 }
 
 /// Read the Nitrokey configuration.
-fn config_get(args: Vec<String>) -> Result<()> {
+fn config_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Prints the Nitrokey configuration");
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
 
-  commands::config_get()
+  commands::config_get(ctx)
 }
 
 /// Write the Nitrokey configuration.
-fn config_set(args: Vec<String>) -> Result<()> {
+fn config_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut numlock = None;
   let mut no_numlock = false;
   let mut capslock = None;
@@ -573,7 +390,7 @@ fn config_set(args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Allow one-time password generation without PIN",
   );
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
   let numlock = ConfigOption::try_from(no_numlock, numlock, "numlock")?;
@@ -586,20 +403,20 @@ fn config_set(args: Vec<String>) -> Result<()> {
   } else {
     None
   };
-  commands::config_set(numlock, capslock, scrollock, otp_pin)
+  commands::config_set(ctx, numlock, capslock, scrollock, otp_pin)
 }
 
 /// Lock the Nitrokey.
-fn lock(args: Vec<String>) -> Result<()> {
+fn lock(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Locks the connected Nitrokey device");
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
 
-  commands::lock()
+  commands::lock(ctx)
 }
 
 /// Execute an OTP subcommand.
-fn otp(args: Vec<String>) -> Result<()> {
+fn otp(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut subcommand = OtpCommand::Get;
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
@@ -615,15 +432,15 @@ fn otp(args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  subargs.insert(0, format!("nitrocli otp {}", subcommand));
-  subcommand.execute(subargs)
+  subargs.insert(0, format!("nitrocli {} {}", Command::Otp, subcommand));
+  subcommand.execute(ctx, subargs)
 }
 
 /// Generate a one-time password on the Nitrokey device.
-fn otp_get(args: Vec<String>) -> Result<()> {
+fn otp_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut slot: u8 = 0;
   let mut algorithm = OtpAlgorithm::Totp;
   let mut time: Option<u64> = None;
@@ -644,14 +461,14 @@ fn otp_get(args: Vec<String>) -> Result<()> {
     argparse::StoreOption,
     "The time to use for TOTP generation (Unix timestamp, default: system time)",
   );
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  commands::otp_get(slot, algorithm, time)
+  commands::otp_get(ctx, slot, algorithm, time)
 }
 
 /// Configure a one-time password slot on the Nitrokey device.
-pub fn otp_set(args: Vec<String>) -> Result<()> {
+pub fn otp_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut slot: u8 = 0;
   let mut algorithm = OtpAlgorithm::Totp;
   let mut name = "".to_owned();
@@ -660,6 +477,7 @@ pub fn otp_set(args: Vec<String>) -> Result<()> {
   let mut counter: u64 = 0;
   let mut time_window: u16 = 30;
   let mut ascii = false;
+  let mut secret_format: Option<OtpSecretFormat> = None;
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Configures a one-time password slot");
   let _ =
@@ -670,7 +488,7 @@ pub fn otp_set(args: Vec<String>) -> Result<()> {
   let _ = parser.refer(&mut algorithm).add_option(
     &["-a", "--algorithm"],
     argparse::Store,
-    "The OTP algorithm to use (hotp or totp, default: totp",
+    "The OTP algorithm to use (hotp or totp, default: totp)",
   );
   let _ = parser.refer(&mut name).required().add_argument(
     "name",
@@ -700,11 +518,31 @@ pub fn otp_set(args: Vec<String>) -> Result<()> {
   let _ = parser.refer(&mut ascii).add_option(
     &["--ascii"],
     argparse::StoreTrue,
-    "Interpret the given secret as an ASCII string of the secret",
+    "Interpret the given secret as an ASCII string of the secret (deprecated, use --format instead)"
   );
-  parse(&parser, args)?;
+  let _ = parser.refer(&mut secret_format).add_option(
+    &["-f", "--format"],
+    argparse::StoreOption,
+    "The format of the secret (ascii|base32|hex)",
+  );
+  parse(ctx, &parser, args)?;
   drop(parser);
 
+  if ascii {
+    if secret_format.is_some() {
+      return Err(Error::Error(
+        "The --format and the --ascii option cannot be used at the same time".to_string(),
+      ));
+    }
+
+    println!(
+      ctx,
+      "Warning: The --ascii option is deprecated. Please use --format ascii instead."
+    )?;
+    secret_format = Some(OtpSecretFormat::Ascii);
+  }
+  let secret_format = secret_format.unwrap_or(OtpSecretFormat::Hex);
+
   let data = nitrokey::OtpSlotData {
     number: slot,
     name,
@@ -713,11 +551,11 @@ pub fn otp_set(args: Vec<String>) -> Result<()> {
     use_enter: false,
     token_id: None,
   };
-  commands::otp_set(data, algorithm, counter, time_window, ascii)
+  commands::otp_set(ctx, data, algorithm, counter, time_window, secret_format)
 }
 
 /// Clear an OTP slot.
-fn otp_clear(args: Vec<String>) -> Result<()> {
+fn otp_clear(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut slot: u8 = 0;
   let mut algorithm = OtpAlgorithm::Totp;
   let mut parser = argparse::ArgumentParser::new();
@@ -732,14 +570,14 @@ fn otp_clear(args: Vec<String>) -> Result<()> {
     argparse::Store,
     "The OTP algorithm to use (hotp|totp)",
   );
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  commands::otp_clear(slot, algorithm)
+  commands::otp_clear(ctx, slot, algorithm)
 }
 
 /// Print the status of the OTP slots.
-fn otp_status(args: Vec<String>) -> Result<()> {
+fn otp_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut all = false;
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Prints the status of the OTP slots");
@@ -748,14 +586,14 @@ fn otp_status(args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Show slots that are not programmed",
   );
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  commands::otp_status(all)
+  commands::otp_status(ctx, all)
 }
 
 /// Execute a PIN subcommand.
-fn pin(args: Vec<String>) -> Result<()> {
+fn pin(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut subcommand = PinCommand::Clear;
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
@@ -771,24 +609,24 @@ fn pin(args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  subargs.insert(0, format!("nitrocli pin {}", subcommand));
-  subcommand.execute(subargs)
+  subargs.insert(0, format!("nitrocli {} {}", Command::Pin, subcommand));
+  subcommand.execute(ctx, subargs)
 }
 
 /// Clear the PIN as cached by various other commands.
-fn pin_clear(args: Vec<String>) -> Result<()> {
+fn pin_clear(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Clears the cached PINs");
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
 
   commands::pin_clear()
 }
 
 /// Change a PIN.
-fn pin_set(args: Vec<String>) -> Result<()> {
+fn pin_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut pintype = pinentry::PinType::User;
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Changes a PIN");
@@ -797,23 +635,23 @@ fn pin_set(args: Vec<String>) -> Result<()> {
     argparse::Store,
     "The PIN type to change (admin|user)",
   );
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  commands::pin_set(pintype)
+  commands::pin_set(ctx, pintype)
 }
 
 /// Unblock and reset the user PIN.
-fn pin_unblock(args: Vec<String>) -> Result<()> {
+fn pin_unblock(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Unblocks and resets the user PIN");
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
 
-  commands::pin_unblock()
+  commands::pin_unblock(ctx)
 }
 
 /// Execute a PWS subcommand.
-fn pws(args: Vec<String>) -> Result<()> {
+fn pws(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut subcommand = PwsCommand::Get;
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
@@ -829,15 +667,15 @@ fn pws(args: Vec<String>) -> Result<()> {
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  subargs.insert(0, format!("nitrocli pws {}", subcommand));
-  subcommand.execute(subargs)
+  subargs.insert(0, format!("nitrocli {} {}", Command::Pws, subcommand));
+  subcommand.execute(ctx, subargs)
 }
 
 /// Access a slot of the password safe on the Nitrokey.
-fn pws_get(args: Vec<String>) -> Result<()> {
+fn pws_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut slot: u8 = 0;
   let mut name = false;
   let mut login = false;
@@ -870,14 +708,14 @@ fn pws_get(args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Print the stored data without description",
   );
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  commands::pws_get(slot, name, login, password, quiet)
+  commands::pws_get(ctx, slot, name, login, password, quiet)
 }
 
 /// Set a slot of the password safe on the Nitrokey.
-fn pws_set(args: Vec<String>) -> Result<()> {
+fn pws_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut slot: u8 = 0;
   let mut name = String::new();
   let mut login = String::new();
@@ -904,14 +742,14 @@ fn pws_set(args: Vec<String>) -> Result<()> {
     argparse::Store,
     "The password to store on the slot",
   );
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  commands::pws_set(slot, &name, &login, &password)
+  commands::pws_set(ctx, slot, &name, &login, &password)
 }
 
 /// Clear a PWS slot.
-fn pws_clear(args: Vec<String>) -> Result<()> {
+fn pws_clear(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut slot: u8 = 0;
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Clears a password safe slot");
@@ -920,14 +758,14 @@ fn pws_clear(args: Vec<String>) -> Result<()> {
     argparse::Store,
     "The PWS slot to clear",
   );
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  commands::pws_clear(slot)
+  commands::pws_clear(ctx, slot)
 }
 
 /// Print the status of the PWS slots.
-fn pws_status(args: Vec<String>) -> Result<()> {
+fn pws_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut all = false;
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Prints the status of the PWS slots");
@@ -936,18 +774,33 @@ fn pws_status(args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Show slots that are not programmed",
   );
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
-  commands::pws_status(all)
+  commands::pws_status(ctx, all)
 }
 
 /// Parse the command-line arguments and return the selected command and
 /// the remaining arguments for the command.
-fn parse_arguments(args: Vec<String>) -> Result<(Command, Vec<String>)> {
+fn parse_arguments<'io, 'ctx: 'io>(
+  ctx: &'ctx mut RunCtx<'_>,
+  args: Vec<String>,
+) -> Result<(Command, ExecCtx<'io>, Vec<String>)> {
+  let mut model: Option<DeviceModel> = None;
+  let mut verbosity = 0;
   let mut command = Command::Status;
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
+  let _ = parser.refer(&mut verbosity).add_option(
+    &["-v", "--verbose"],
+    argparse::IncrBy::<u64>(1),
+    "Increase the log level (can be supplied multiple times)",
+  );
+  let _ = parser.refer(&mut model).add_option(
+    &["-m", "--model"],
+    argparse::StoreOption,
+    "Select the device model to connect to (pro|storage)",
+  );
   parser.set_description("Provides access to a Nitrokey device");
   let _ = parser.refer(&mut command).required().add_argument(
     "command",
@@ -960,15 +813,26 @@ fn parse_arguments(args: Vec<String>) -> Result<(Command, Vec<String>)> {
     "The arguments for the command",
   );
   parser.stop_on_first_argument(true);
-  parse(&parser, args)?;
+  parse(ctx, &parser, args)?;
   drop(parser);
 
   subargs.insert(0, format!("nitrocli {}", command));
-  Ok((command, subargs))
+
+  let ctx = ExecCtx {
+    model,
+    stdout: ctx.stdout,
+    stderr: ctx.stderr,
+    admin_pin: ctx.admin_pin.take(),
+    user_pin: ctx.user_pin.take(),
+    new_admin_pin: ctx.new_admin_pin.take(),
+    new_user_pin: ctx.new_user_pin.take(),
+    verbosity,
+  };
+  Ok((command, ctx, subargs))
 }
 
 /// Parse the command-line arguments and execute the selected command.
-pub fn handle_arguments(args: Vec<String>) -> Result<()> {
-  let (command, args) = parse_arguments(args)?;
-  command.execute(args)
+pub(crate) fn handle_arguments(ctx: &mut RunCtx<'_>, args: Vec<String>) -> Result<()> {
+  let (command, mut ctx, args) = parse_arguments(ctx, args)?;
+  command.execute(&mut ctx, args)
 }
